/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <stdio.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <stdarg.h>
#include <syslog.h>
#include <strings.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/time.h>

#include <pth.h>

#include "libxode.h"

#ifndef _ETHERX_H_
#define _ETHERX_H_

extern int etherx_debug_flag;
#define log_debug if(etherx_debug_flag) _log_debug

extern unsigned int etherx_stack_default;

int log_init(int log2, char *arg);
void etherx_debug();


/* this is a special data type, a thread "handle", that stores thread-lifetime data */

typedef struct thread_struct
{
    pth_t id;
    pool p;
    void *data;
} _thread, *thread;

thread thread_new(); /* creates new thread handle */
thread thread_get(); /* gets the current thread's handle */
void thread_free(thread t); /* frees a thread handle */

extern int _tlisten_default;

typedef void (*tlisten_onConnect)(thread t, int sock, struct sockaddr_in sa);

typedef struct tlisten
{
     int                         s;	/* Base accept/server socket */
     tlisten_onConnect connect_evt;	/* Function pointer for onConnect event */
     void        *connect_evt_data;	/* Used by onConnect event */
     char                      *ip;
     u_short                  port;
} _tlisten, *tlisten;

/* Creates the tlisten stuct, socket, and thread to accept connects */
thread tlisten_new(u_short port, char *ip, tlisten_onConnect evt, void *evt_data);

/* This is the main loop for the tlisten thread */
void *_tlisten_main(void *arg);

/* Called when the thread exits to clean up any tlisten data and free the tlisten struct */
void _tlisten_cleanup(void *arg);

typedef struct tstream_buffer_struct
{
    pth_message_t head; /* the standard pth message header */
    char *data;
    int data_sz;
    int sent;
    void *obj; /* can point to the obj that created the data being sent, in the same pool */
    struct tstream_buffer_struct *next;
    pool p;
} *tstream_buffer;

typedef void (*tstream_handler_read)(thread t, void *buffer, int bufsz);
typedef void (*tstream_handler_err)(thread t, tstream_buffer buffer);

typedef struct tstream_struct
{
    int fd, port, close;
    char *host;

    tstream_handler_read h_read; /* Function pointer for handling read data */
    tstream_handler_err h_err; /* Handles data left in the write buffer when the thread gets killed */
    void *data, *data1, *data2, *data3, *data4, *data5; /* Used by the handlers/internally */
    int flag, flag1, flag2, flag3, flag4, flag5;

    void *data_aux1, *data_aux2, *data_aux3, *data_aux4, *data_aux5; /* For general purpose usage within thread */
    int flag_aux1, flag_aux2, flag_aux3, flag_aux4, flag_aux5;

    /* actual calls */
    int (*reader)(struct tstream_struct *ts, void *buf, size_t count);
    int (*writer)(struct tstream_struct *ts, const void *buf, size_t count);

    tstream_buffer write; /* The current data being written */
    tstream_buffer oob; /* out of band data, waiting to be written */
    pth_msgport_t write_mp; /* Pth message port for unwritten blocked data */

} _tstream, *tstream;

/* starts tstream on open descriptor */
thread tstream_new(int fd, char *host, tstream_handler_read read, tstream_handler_err err, void *data);

/* uses adns if available to resolve the host and connect to the port
 * attempts to connect to port on A first, failing that, does a reverse-MX sort lookup attempting to connect to each entry
 */
thread tstream_connect(int port, char *host, tstream_handler_read read, tstream_handler_err err, void *data);

/* Specifies new data to be written to the fd */
void tstream_write(thread t, void *buffer, int bufsz);
void tstream_write_buffer(thread t, tstream_buffer buffer);

/* out of band data, don't write it until specifically flushed */
void tstream_oob_buffer(thread t, tstream_buffer buffer);
void tstream_oob_flush(thread t);

/* close the thread nicely, after all the data was written */
void tstream_close(thread t);

/* thread internal utils */
void *_tstream_main(void *arg);
thread _tstream_new(int fd, int port, char *host, tstream_handler_read read, tstream_handler_err err, void *data);
int _tstream_connect(int port, char *host);
void _tstream_write(thread t);
void _tstream_cleanup(void *arg);
void _tstream_cleanup_evt(void *arg);
int _tstream_socket(int port, struct in_addr *iap);
void _tstream_cpu(thread t);


#define XMLSTREAM_ERROR -1
#define XMLSTREAM_ROOT 0
#define XMLSTREAM_NODE 1
#define XMLSTREAM_CLOSE 2
#define XMLSTREAM_INIT 3
#define XMLSTREAM_BOUNCE 4

#define XML_MAX_DEPTH 100

typedef int (*xmlstream_onNode)(thread t, xmlnode x, int type);

/* Wrapper around tstream_* */
thread xmlstream_new(int fd, char *host, xmlstream_onNode evt);
thread xmlstream_connect(int port, char *host, xmlstream_onNode evt);

/* Specifies a new xmlnode to serialize to the fd */
int xmlstream_write(thread t, xmlnode x);

/* Special utilities */
xmlnode xmlstream_header(char *namespace, char *to, char *from);
void xmlstream_write_header(thread t, xmlnode x);
void xmlstream_close(thread t);
void xmlstream_error(thread t, char *err); /* calls xmlstream_close() */

/* The onRead callback for the tstream */
void _xmlstream_main(thread t, void *buffer, int bufsz);
void _xmlstream_error(thread t, tstream_buffer buffer);

void _xmlstream_cleanup(void *arg);


#define LOGNONE	0
#define LOGSTDERR	1
#define LOGFILE	2
#define LOGSYSLOG	3

#define LOGSIZE_HDR  1024 + 1
#define LOGSIZE_TAIL 2048 + 1

#define ZONE zonestr(__FILE__,__LINE__)

extern int log_device;
extern thread log_device_file;

void _log_debug(char *zone, const char *msgfmt, ...);
void _log(const char *logmsg);
void log_warn(char *zone, const char *msgfmt, ...);
void log_error(char *zone, const char *msgfmt, ...);
char *create_log_timestamp(void);
char *zonestr(char *file, int line);


typedef struct rlimit_struct
{
    char *key;
    int start;
    int points;
    int maxt, maxp;
    pool p;
} *rlimit, _rlimit;
 
rlimit rate_new(int maxt, int maxp);
void rate_free(rlimit r);
int rate_check(rlimit r, char *key, int points);



#define ETHERX_NAMESPACE "http://etherx.jabber.org/"

#define EPACKET_UNKNOWN -1
#define EPACKET_NORMAL 0 /* default */
#define EPACKET_ERROR 1 /* special flag that this is an error/bounced packet */

#define EHANDLER_ERROR 0
#define EHANDLER_DAEMON 1
#define EHANDLER_SOCKET 2


typedef struct epacket_struct
{
    xmlnode x;
    char *to;
    char *from;
    char *ns;
    char *error;
    int type;
} _epacket, *epacket;

typedef void (*etherx_handler)(epacket e);

typedef struct ehandler_struct
{
    int type;
    char *host;
    char *namespace;
    etherx_handler h;
    pth_mutex_t mutex;
    pth_cond_t cond;
    thread t;
    pool p;

    /* these are used within etherxd */
    struct ehandler_struct *next, *children;
    int count;

} _ehandler, *ehandler;

/* public functions */
void etherx_init(char *secret, char *host); /* transport init */
void ehandler_send(ehandler eh, xmlnode x, char *to); /* tells etherx (via ehandler) to send x to to */
ehandler ehandler_new(etherx_handler h, char *host, char *namespace); /* creates a new ehandler for this host/namespace */
void epacket_error(epacket e, char *error); /* simply turns a normal epacket into an error one */
epacket epacket_new(xmlnode x, char *to, char *from, char *ns); /* if to/from are null it extracts them from x, sets type based on element name */

xmlnode epacket_serialize(epacket e); /* puts etherx:* attribs into the xml */
void epacket_deserialize(epacket e); /* puts etherx:* attribs into the epacket and removes them from the xml */

/* list utils */
ehandler ehandler_list_add(ehandler list, ehandler item);
ehandler ehandler_list_isolate(ehandler list, ehandler item);
ehandler ehandler_list_free(ehandler list, ehandler item);
ehandler ehandler_list_get(ehandler list, char *host, char *namespace);

#endif
