/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Jabber
 *  Copyright (C) 1998-1999 The Jabber Team http://jabber.org/
 */

#include <etherxd.h>

char *etherx__host = NULL;
char *etherx__secret = NULL;
int etherx__daemon = 0;

void etherx_init(char *secret, char *host)
{
    etherx__host = strdup(host);
    etherx__secret = strdup(secret);
    log_debug(ZONE,"etherx_init(%s,%s)",secret,host);
}

void ehandler_connect(ehandler eh)
{ /* we block until success or failure */
    log_debug(ZONE,"ehandler_connect!");

    /* if we started a working daemon already */
    if(etherx__daemon)
    {
        etherxd_new(eh); /* tell it about a new transport handler */
        return;
    }

    /* attempt an xmlstream, unlocks when connected or error */
    etherxs_connect(eh);

    log_debug(ZONE,"spun off connection attempt");

    /* wait for lock from stream connect, and then we don't need it anymore */
    pth_mutex_acquire(&(eh->mutex), FALSE, NULL);
    pth_cond_await(&(eh->cond), &(eh->mutex), NULL);
    log_debug(ZONE,"condition happened! eh->type is %d",eh->type);
    pth_mutex_release(&(eh->mutex));

    /* we're connected as a stream */
    if(eh->type != EHANDLER_ERROR)
        return;

    log_debug(ZONE,"no luck connecting");

    /* attempt a listen daemon */
    etherxd_new(eh);

    /* if it's running, flag for future use */
    if(eh->type == EHANDLER_DAEMON)
        etherx__daemon = 1;

    log_debug(ZONE,"returning eh->type %d",eh->type);
}

ehandler ehandler_new(etherx_handler h, char *host, char *namespace)
{
    pool p;
    ehandler e;

    log_debug(ZONE,"ehandler_new(%s,%s)",host,namespace);

    /* init new ehandler */
    p = pool_new();
    e = pmalloc(p, sizeof(_ehandler));
    e->host = pstrdup(p, host);
    e->namespace = pstrdup(p, namespace);
    e->p = p;
    e->h = h;
    e->t = NULL;
    e->type = EHANDLER_ERROR;

    /* we're not a true handler, just a stub */
    if(h == NULL)
        return e;

    /* init the mutex and try connecting */
    if(pth_mutex_init(&(e->mutex)) && pth_cond_init(&(e->cond)))
        ehandler_connect(e);

    if(e->type != EHANDLER_ERROR)
        return e;

    log_error("libetherx","Failed to connect or become daemon");
    pool_free(p);
    return NULL;
}

void ehandler_send(ehandler eh, xmlnode x, char *to)
{
    epacket ep;

    if(eh == NULL || x == NULL || to == NULL)
    {
        log_warn("libetherx","received illegal outgoing data");
        return;
    }

    ep = epacket_new(x, to, eh->host, eh->namespace);

    if(etherx__daemon)
    {
        log_debug(ZONE,"ehandler_send: daemon %s",xmlnode2str(x));
        etherxd_send(ep);
        return;
    }
    if(eh->type == EHANDLER_SOCKET)
    {
        log_debug(ZONE,"ehandler_send: socket %s",xmlnode2str(x));
        xmlstream_write(eh->t, epacket_serialize(ep));
        return;
    }

    /* if we get here it's because the type is EHANDLER_ERROR from the socket disconnecting, bounce back to the handler */
    epacket_error(ep,"delivery failure");
    (*eh->h)(ep);
}

